//Debug output router
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "DebugRouter.h"
#include "Setup.h"
#include "MemMan.h"

#include <iostream> //for RouterOutputStdout
#include <algorithm>

namespace MPMA
{
	//the global framework error reporter
	RouterInput *errorReportPointer=0;
	
	//init stuff
	namespace
	{
		void ReportsInitialize()
		{
			errorReportPointer=new2(RouterInput(),RouterInput);
		}
		void ReportsShutdown()
		{
			delete2(errorReportPointer);
			errorReportPointer=0;
		}
		
		class AutoInitReports
		{
		public:
			//hookup init callbacks
			AutoInitReports()
			{
				MPMA::Internal_AddInitCallback(ReportsInitialize,-1000);
				MPMA::Internal_AddShutdownCallback(ReportsShutdown,-1000);
			}
		} autoInitReports;
	}
		
	RouterOutput::~RouterOutput()
	{
		TakeSpinLock takeLock(feedLock);
		for (std::list<RouterInput*>::iterator i=reportFeeds.begin(); i!=reportFeeds.end(); ++i)
			(**i).RemoveOutputMethod(this, false, false);
	}
	
	void RouterOutput::Output(const unsigned char *data, unsigned int dataLen)
	{
		ErrorReport()<<"Data is being written to a base class version of RouterOutput.  This indicates that either a derived class did not implement Output, or that the derived class has been destructed already, but something tried to write output to it.  Data len="<<dataLen<<"\n";
	}
	
	void RouterOutput::DetachFeed(RouterInput *feed)
	{
		TakeSpinLock takeLock(feedLock);
		feed->RemoveOutputMethod(this, false, false);
		reportFeeds.remove(feed);
	}

	void RouterOutput::FlushInputSources()
	{
		TakeSpinLock takeLock(feedLock);
		for (std::list<RouterInput*>::iterator i=reportFeeds.begin(); i!=reportFeeds.end(); ++i)
			(**i).FlushInputSources();
	}
	
	// --
#ifdef DEBUGROUTER_ENABLED	
	RouterInput::RouterInput()
	{
		bufferedDataProcessing=false;
		numNonbufferedOutputs=0;
		
		//set us up the buffer
		buffer=new2_array(unsigned char, DEBUGROUTER_BUFFER_SIZE, unsigned char);
		bufferLen=0;
		
		//create the data marshelling worker thread
		destructing=false;
		dataMover=new2(Thread(ThreadProc,this),Thread);
		
		alive=true;
	}
	
	RouterInput::~RouterInput()
	{
		alive=false;
		
		//stop/free the thread
		destructing=true;
		workerWaiter.Clear();
		delete2(dataMover);

		//make sure that our data actually all made it out
		FlushOutput();
		
		//remove ourself from all outputs's lists
		int outputCount=outputs.size();
		while (!outputs.empty())
		{
			outputs.front()->DetachFeed(this);
		}
		
		//
		if (numNonbufferedOutputs!=0)
		{
			ErrorReport()<<"numNonbufferedOutputs is not 0 on destruction of a RouterInput object it was "<<numNonbufferedOutputs<<", had "<<outputCount<<" outputs before destruction and "<<outputs.size()<<" after.\n";
		}
		
		outputs.clear();
		
		//set us down the buffer
		delete2_array(buffer);
		buffer=0;
		bufferLen=0;
	}

	//writes all pending data to all destinations
	void RouterInput::FlushOutput()
	{
		if (bufferLen==0) return;

		TakeMutexLock takelock(outputLock);
		for (std::list<RouterOutput*>::iterator i=outputs.begin(); i!=outputs.end(); ++i)
		{
			if ((**i).IsBufferProcessing())
				(**i).Output((unsigned char*)buffer, bufferLen);
		}

		bufferLen=0;
	}

	//
	void RouterInput::FlushInputSources()
	{
		TakeSpinLock takethreadworklock(threadWorkingLock);
		RouterOutput::FlushInputSources();
		FlushOutput();
	}
	
	//adds an output destination to this input
	void RouterInput::AddOutputMethod(RouterOutput *outputter)
	{
		if (!alive) return;
		
		TakeMutexLock takelock(outputLock);
		outputs.push_front(outputter);
		
		if (!outputter->IsBufferProcessing()) ++numNonbufferedOutputs;

		TakeSpinLock takeLock(outputter->feedLock);
		outputter->reportFeeds.push_back(this);
	}
	
	//removes an output destination from this input
	void RouterInput::RemoveOutputMethod(RouterOutput *outputter, bool flushOutputFirst, bool removeChildLink)
	{
		TakeMutexLock takelock(outputLock);

		//make sure all output is flushed first so nothing is lost
		if (flushOutputFirst)
			FlushOutput();

		//do remove
		TakeSpinLock takethreadworklock(threadWorkingLock);
		if (std::find(outputs.begin(), outputs.end(), outputter)!=outputs.end())
		{
			outputs.remove(outputter);
	
			if (!outputter->IsBufferProcessing()) --numNonbufferedOutputs;
			
			if (removeChildLink)
			{
				TakeSpinLock takeLock(outputter->feedLock);
				outputter->reportFeeds.remove(this);
			}
		}
	}
	
	//handles input to the router and moves it to the output queue
	void RouterInput::Output(const unsigned char *data, unsigned int dataLen)
	{
		if (!alive) return;
		if (dataLen==0) return;
		if (outputs.size()==0) return;
		
		TakeMutexLock lockMultiCalls(addDataLock);
		
		//do nonbuffered ones
		if (numNonbufferedOutputs>0)
		{
			TakeMutexLock takelock(outputLock);
			
			for (std::list<RouterOutput*>::iterator i=outputs.begin(); i!=outputs.end(); ++i)
			{
				if (!(**i).IsBufferProcessing())
					(**i).Output(data, dataLen);
			}
		}
		
		if (outputs.size()==numNonbufferedOutputs) return; //no more to do
		
		//now do buffered ones
		TakeSpinLock takelock(dataLock);
		
		unsigned int dataOffset=0;
		unsigned int dataLeft=dataLen;
		while (dataLeft>0)
		{
			int amt=DEBUGROUTER_BUFFER_SIZE-bufferLen;
			if (amt>(int)dataLeft) amt=dataLeft;
			
			memcpy((void*)(buffer+bufferLen), data+dataOffset, amt);
			bufferLen+=amt;
			dataOffset+=amt;
			dataLeft-=amt;

			if (dataLeft>0) //need to wait on thread
			{
				routerWaiter.Set();
				takelock.Leave();
				workerWaiter.Clear();
				routerWaiter.WaitUntilClear();
				takelock.Take();
			}
			else //else just wake it up
				workerWaiter.Clear();
		}
	}
	
	//data marshelling thread
	void RouterInput::ThreadProc(Thread& thread, ThreadParam param)
	{
		RouterInput *me=(RouterInput*)param.ptr;
		
		unsigned char *localData=new2_array(unsigned char, DEBUGROUTER_BUFFER_SIZE, unsigned char);
		unsigned int localDataLen=0;
		
		while (!thread.IsEnding() && !me->destructing)
		{
			me->workerWaiter.Set();
			{
				TakeSpinLock takethreadworklock(me->threadWorkingLock);
				//swap the local buffer with the main one
				{
					TakeSpinLock takelock(me->dataLock);
					
					unsigned char *tmp0=localData;
					localData=(unsigned char*)me->buffer;
					me->buffer=tmp0;
					
					unsigned int tmp1=localDataLen;
					localDataLen=me->bufferLen;
					me->bufferLen=tmp1;
					
					me->routerWaiter.Clear();
				}
				
				//send to all outputs
				if (localDataLen>0)
				{
					TakeMutexLock takelock(me->outputLock);

					for (std::list<RouterOutput*>::iterator i=me->outputs.begin(); i!=me->outputs.end(); ++i)
					{
						if ((**i).IsBufferProcessing())
							(**i).Output(localData, localDataLen);
					}
					
					localDataLen=0;
				}
			}
			me->workerWaiter.WaitUntilClear();
		}
		
		delete2_array(localData);
	}
#endif
	
	// -- useful outputs
	
	//stdout
	RouterOutputStdout::~RouterOutputStdout()
	{
		FlushInputSources();
	}

	void RouterOutputStdout::Output(const unsigned char *data, unsigned int dataLen)
	{
#ifdef DEBUGROUTER_ENABLED
		std::cout.write((const char*)data, dataLen);
#endif
	}
	
	//file
	RouterOutputFile::RouterOutputFile(const Filename &fileName, bool textMode)
	{
#ifdef DEBUGROUTER_ENABLED
		f=fopen(fileName.c_str(), textMode?"wt":"wb");
#endif
	}
	
	RouterOutputFile::~RouterOutputFile()
	{
#ifdef DEBUGROUTER_ENABLED
		FlushInputSources();
		TakeMutexLock take(lock);
		if (f) fclose(f);
		f=0;
#endif
	}
	
	void RouterOutputFile::Output(const unsigned char *data, unsigned int dataLen)
	{
#ifdef DEBUGROUTER_ENABLED
		if (!f) return;
		TakeMutexLock take(lock);
		if (!f) return;
		fwrite(data,dataLen,1,f);
		fflush(f);
#endif
	}
}
