//smart rw thread synchronizator.  read-only accesses don't block each other out.
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "Locks.h"
#include "DebugRouter.h"
#include "Debug.h"
#include "Thread.h"
#include "Types.h"
#include "Info.h"
#include <list>
#include <algorithm>

namespace MPMA
{

// -- RWSleepLock

//
RWSleepLock::RWSleepLock()
{
	cntRead=cntWrite=0;
	cntWritePend=cntReadPend=0;
}

//
RWSleepLock::~RWSleepLock()
{
#ifdef _DEBUG
	if (cntRead || cntWrite || cntWritePend || cntReadPend)
	{
		ErrorReport()<<"BAD: a RWSleepLock is destructing, but some TakeRWSleepLock is still using it.\n";
		ErrorReport()<<"  -call stack: "<<MPMA::GetCallStack()<<"\n";
	}
#endif
}

namespace
{
	THREAD_LOCAL std::list<RWSleepLock*> *myThreadLocksRW=0; //list of locks held by this thread
	
	//acquires access to a RWSync section (blocks if needed) by manipulating the read/write counters
	inline void AcquireRead(volatile unsigned int &rc,volatile  unsigned int &wc,volatile  unsigned int &rpend,volatile  unsigned int &wpend, RWSleepLock *who)
	{
		//init thread info if we've never done before
		if (!myThreadLocksRW)
			myThreadLocksRW=new std::list<RWSleepLock*>();

		//if it's OUR thread holding this lock already, then it's ok to take
		if (std::find(myThreadLocksRW->begin(),myThreadLocksRW->end(),who)!=myThreadLocksRW->end())
		{
			AtomicIntInc(&rc);

			myThreadLocksRW->insert(myThreadLocksRW->begin(),who);
			return;
		}
		
		//mark us as pending a read
		AtomicIntInc(&rpend);

		//pending writes go before reads
		while (wpend)
		{
			//else just wait and try later
			Sleep(0);
		}

		//
		while (true)
		{
			//wait until there is no write access, then grab the writer lock
			while (wc)
				Sleep(0);
			
			unsigned int result;
			if (!AtomicCompareExchange(&wc, 0, 1, result))
				continue;
			
			//we got the writer, so take the reader then release writer
			AtomicIntInc(&rc);
			AtomicIntDec(&wc);

			break;
		}

		//
		myThreadLocksRW->insert(myThreadLocksRW->begin(),who);
		AtomicIntDec(&rpend); //release our pending mark
	}
	
	inline void AcquireWrite(volatile unsigned int &rc,volatile  unsigned int &wc,volatile  unsigned int &rpend,volatile  unsigned int &wpend, RWSleepLock *who)
	{
		//init thread info if we've never done before
		if (!myThreadLocksRW)
			myThreadLocksRW=new std::list<RWSleepLock*>();

		//if it's OUR thread holding this lock already, then it's ok to take
		if (std::find(myThreadLocksRW->begin(),myThreadLocksRW->end(),who)!=myThreadLocksRW->end())
		{
			AtomicIntInc(&wc);

			myThreadLocksRW->insert(myThreadLocksRW->begin(),who);
			return;
		}

		//mark us as pending a write
		AtomicIntInc(&wpend);

		//
		while (true)
		{
			//wait until there is no other write access, then try to assert our write access
			while (wc || rc)
				Sleep(0);
			
			unsigned int result;
			if (!AtomicCompareExchange(&wc, 0, 1, result))
				continue;
			
			//wait on any ninja readers...
			while (rc)
				Sleep(0);
			break;
		}

		//
		myThreadLocksRW->insert(myThreadLocksRW->begin(),who);
		AtomicIntDec(&wpend); //release our pending mark
	}
	//releases access to a RWSync section
	inline void ReleaseRead(unsigned int &rc, unsigned int &wc, unsigned int &rpend, unsigned int &wpend, RWSleepLock *who)
	{
		AtomicIntDec(&rc);
		myThreadLocksRW->erase(std::find(myThreadLocksRW->begin(),myThreadLocksRW->end(),who));
	}
	inline void ReleaseWrite(unsigned int &rc, unsigned int &wc, unsigned int &rpend, unsigned int &wpend, RWSleepLock *who)
	{
		AtomicIntDec(&wc);
		myThreadLocksRW->erase(std::find(myThreadLocksRW->begin(),myThreadLocksRW->end(),who));

		if (wpend || rpend) //if we had caused something else to wait, give them a chance
			Sleep(0);
	}
}

//take or leave the RWSync section
void TakeRWSleepLock::Take(bool writeAccess)
{
	//if already taken...
	if (meIn)
	{
		if (writeAccess==meWrite) return; //redundant request...
		else Leave(); //changing access type, so give up old
	}

	//take it
	meIn=true;
	meWrite=writeAccess;

	if (meWrite) AcquireWrite(crit.cntRead, crit.cntWrite, crit.cntReadPend, crit.cntWritePend, &crit);
	else AcquireRead(crit.cntRead, crit.cntWrite, crit.cntReadPend, crit.cntWritePend, &crit);

}
void TakeRWSleepLock::Leave()
{
	if (meIn)
	{
		meIn=false;
		if (meWrite) ReleaseWrite(crit.cntRead, crit.cntWrite, crit.cntReadPend, crit.cntWritePend, &crit);
		else ReleaseRead(crit.cntRead, crit.cntWrite, crit.cntReadPend, crit.cntWritePend, &crit);
	}
}

// -- Spinlock

void TakeSpinLock::Take()
{
	unsigned int yawn;
	while (!AtomicCompareExchange(&locker.taken, 0, 1, yawn))
	{
		locker.collision=true;
		if (SystemInfo::SuggestSleepInSpinlock)
			Sleep(0);
	}
	
	taken=true;
}

void TakeSpinLock::Leave()
{
#ifdef _DEBUG
	if (locker.taken!=1 || !taken)
	{
		ErrorReport()<<"Spinlock is being left but was not taken.\n";
	}
#endif
	
	bool wasCollision=locker.collision;
	locker.collision=false;
	locker.taken=0;
	if (wasCollision && SystemInfo::SuggestSleepInSpinlock)
	{
		Sleep(0);
	}
	
	taken=false;
}

} //namespace MPMA
