//!\file base/Locks.h Thread locking constructs and atomic functions.
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#pragma once

namespace MPMA
{

	//  -- Atomic Operations safe for multiprocessor operations
	
	//!Atomically increments an integer.
	void AtomicIntInc(volatile unsigned int *pint);
	
	//!Atomically decrements an integer.
	void AtomicIntDec(volatile unsigned int *pint);
	
	//!Atomically adds one integer to another and returns the value of the original.
	int AtomicIntAdd(volatile int *pint, int addValue);
	
	//!Compares expectedValue with the value at pInt, and if they are the same, sets pInt to newValue and returns true with outResultValue set to newValue.  If they are different then pInt is unaffected, and returns false with outResultValue set to the value that was found at pInt.
	bool AtomicCompareExchange(volatile unsigned int *pInt, unsigned int expectedValue, unsigned int newValue, volatile unsigned int &outResultValue);
	
	
	// -- Locking constructs
	
	//!\brief ReaderWriter sleep-lock (re-entrant safe).
	//!Read locks don't block each other out, but Write locks are exclusive to all other locks.
	class RWSleepLock
	{
		friend class TakeRWSleepLock;
	
	public:
		RWSleepLock();
		~RWSleepLock();
	
	protected:
		unsigned int cntRead, cntWrite; //lock count for read/write access
	
		unsigned int cntWritePend, cntReadPend; //pending access counter
	};
	
	//!Used to lock/unlock a RWSleepLock automatically using scope.  The lock is released on dustruction.
	class TakeRWSleepLock
	{
	public:
		//!ctor
		inline TakeRWSleepLock(RWSleepLock &rwsSection, bool writeAccess=true, bool takeNow=true): crit((RWSleepLock&)rwsSection), meIn(false)
			{if (takeNow) Take(writeAccess); }
		inline ~TakeRWSleepLock()
			{  if (meIn) Leave(); }
	
		//!Manually takes the lock. (safe to call repeatedly)
		void Take(bool writeAccess);
		//!Manually releases the lock. (safe to call repeatedly)
		void Leave();
	
	private:
		mutable RWSleepLock &crit;
	
		bool meIn;
		bool meWrite;
	};
	
	
	//!Mutex (re-entrant safe).
	class MutexLock
	{
		friend class TakeMutexLock;
	
	public:
		MutexLock();
		~MutexLock();
	
	protected:
		void *crit;
		bool isMemManaged; //special case, since memory manager uses us
	};
	
	//!Used to lock/unlock a MutexLock automatically using scope.  The lock is released on dustruction.
	class TakeMutexLock
	{
	public:
		//!ctor
		TakeMutexLock(MutexLock &critSection);
		~TakeMutexLock();
	
	private:
		MutexLock &crit;
	};
	
	
	//!Spinlock (NOT re-entrant safe) that reverts to a sleeplock on single cpu systems.  If no lock-contention occurs, these is extremely little overhead in use.
	class SpinLock
	{
		friend class TakeSpinLock;
	protected:
		volatile unsigned int taken;
		bool collision;
	public:
		inline SpinLock(): taken(0), collision(false) {}
	};
	
	//!Used to lock/unlock a spinlock automatically using scope.  The lock is released on destruction.
	class TakeSpinLock
	{
	public:
		inline TakeSpinLock(SpinLock &slock, bool takeNow=true): locker(slock), taken(false)
			{ if (takeNow) Take(); }
		inline ~TakeSpinLock()
			{ if (taken) Leave(); }
		
		//!Manually takes the lock. (NOT safe to call repeatedly)
		void Take();
		//!Manually releases the lock. (NOT safe to call repeatedly)
		void Leave();
		
	private:
		SpinLock &locker;
		bool taken;
	};
	
	
	// -- Signal-based blocking constructs
	
	//!An object that can be used to block a thread until another thread signals it to resume.
	class BlockingObject
	{
	public:
		BlockingObject();
		~BlockingObject();
		
		//!Marks the object as cleared, so nothing is blocking on it.
		void Clear();
		
		//!Marks the object as set, so BlockUntilClear will block on it.
		void Set();
		
		//!Blocks and waits for as long as the object is set (or if time is non-max, then until that many milliseconds pass).  Returns false only on timeout.
		bool WaitUntilClear(unsigned int timeToWait=0xffffffff);
	
	private:
		SpinLock lock;
		volatile bool isSet;
		void *pdata; //platform-specific container
	};

} //namespace MPMA

#ifdef WIN32
	#include "win32/Locks.h"
#else
	#include "linux/Locks.h"
#endif

