//Luke Lenhart, 2002-2007
//See /docs/License.txt for details on how this code may be used.

#include "Profiler.h"
#include "DebugRouter.h"
#include "Vary.h"

#if defined(_DEBUG) || defined(PROFILE_RELEASE)

const char lpzProfileFile[]="_profile.txt";

//
CProfiler _instProfile; //instance of the profiler

//
CProfiler::CProfiler()
{
	//TODO: Change everything to use the reporter system
	//set up
	FILE *f=fopen(lpzProfileFile,"wt"); //open file and kill old contents
	if (f)
	{
		const char lpzProStartMsg[]="Profile File Started.\n\n";
		fwrite(lpzProStartMsg,strlen(lpzProStartMsg),1,f);

		fclose(f);
	}
	else //can't open..?
	{
		MPMA::ErrorReport()<<"Unable to open file for CProfiler(start).\n";
	}
}

CProfiler::~CProfiler()
{
	FILE* f=fopen(lpzProfileFile,"at");
	if (!f) //couldn't open?
	{
		MPMA::ErrorReport()<<"Unable to open file for CProfiler(end).\n";
	}
	else //it worked, so continue
	{
		double printTime = timeStart.Step();
		Write(f,"Total program execution time: ");
		Write(f,printTime);
		Write(f," seconds.\n\n");

		WriteProfilesToFile(f,printTime);

		fclose(f);
	}

	//dealloc stuff
	DeallocProfiles();
}

//writes a value to a file
void CProfiler::Write(FILE* f, double val)
{
	char buf[64];
	gcvt(val,16,buf);
	fwrite(buf,strlen(buf),1,f);
}

void CProfiler::Write(FILE* f, const char* str)
{
	fwrite(str,strlen(str),1,f);
}

void CProfiler::Write(FILE* f, int num)
{
	std::string buf=(MPMA::Vary)num;
	fwrite(buf.c_str(),buf.size(),1,f);
}

//starts profiling
void CProfiler::_ProfileStart(const std::string &pName, const std::string &fName)
{
	MPMA::TakeMutexLock autoCrit(crit);

	//find profile
	SProfile* pro=FindProfile(pName);
	if (!pro) //profile doesn't exist
	{
		//allocate it
		SProfile newPro;

		//initialize it
		newPro.file=fName;
		newPro.name=pName;
		newPro.samples=0;
		newPro.timeMax=0;
		newPro.timeMin=0;
		newPro.timeTotal=0;
		newPro.started=false;

		//add it to list
		llProf.push_back(newPro);

		pro=&*--llProf.end();
	}
	
	if (pro->started)
	{
		Error("Warning: ProfileStart called without a matching ProfileStop\n");
		Error("  Name: ");
		Error(pName.c_str());
		Error("\n  File: ");
		Error(fName.c_str());
		Error("\n\n");
	}

	//set start
	pro->timeStart.Step();
	pro->started=true;
}

//stops profiling
void CProfiler::_ProfileStop(const std::string &pName, const std::string &fName)
{
	MPMA::Timer searchTimer;

	MPMA::TakeMutexLock autoCrit(crit);

	//find the profile
	SProfile* pro=FindProfile(pName);
	if (!pro || !pro->started) //not found
	{
		Error("Warning: ProfileStop called without a matching ProfileStart\n");
		Error("  Name: ");
		Error(pName.c_str());
		Error("\n  File: ");
		Error(fName.c_str());
		Error("\n\n");
		return;
	}

	//get difference in time
	double timeDif=pro->timeStart.Step();
	pro->started=false;
	
	double searchTime=searchTimer.Step();
	timeDif-=searchTime;

	//update profile appropriatly
	if (timeDif>pro->timeMax) pro->timeMax=timeDif;
	if (timeDif<pro->timeMin || pro->samples==0) pro->timeMin=timeDif;
	pro->timeTotal+=timeDif;
	pro->samples++;
}

//finds a profile, NULL if not found
CProfiler::SProfile* CProfiler::FindProfile(const std::string &name)
{
	std::list<SProfile>::iterator cur=llProf.begin();

	while (cur!=llProf.end())
	{
		if (name==cur->name)
			return &*cur;

		++cur;
	}

	return 0;
}

//writes all profiles taken to file
void CProfiler::WriteProfilesToFile(FILE *f, double finalTime)
{
	//write header
	const char lpzProMsg[]=" -- Profiles --\n\n";
	fwrite(lpzProMsg,strlen(lpzProMsg),1,f);

	//sort by time total
	llProf.sort();

	std::list<SProfile>::iterator cur=llProf.begin();

	//write profiles
	while (cur!=llProf.end())
	{
		Write(f,cur->name.c_str());
		Write(f," - ");
		Write(f,cur->file.c_str());
		Write(f,"\n");

		Write(f," Average time: ");
		Write(f,cur->timeTotal/cur->samples);
		Write(f, " ms\n");

		Write(f," Max time: ");
		Write(f,cur->timeMax);
		Write(f," ms\n");

		Write(f," Min time: ");
		Write(f,cur->timeMin);
		Write(f," ms\n");

		double printTime = cur->timeTotal;
		Write(f," Total time: ");
		Write(f, printTime);
		Write(f," seconds\n");

		Write(f," Samples taken: ");
		Write(f, cur->samples);

		Write(f,"\n Percentage of total time: ");
		Write(f,(int)((printTime/finalTime)*100.0));

		Write(f,"%\n\n");

		++cur;
	}
}

//deallocates strings within profiles, and the profiles themselves
void CProfiler::DeallocProfiles()
{
	//no need for this function with using std::list and std::string
}

//opens profile file and writes error message
void CProfiler::Error(const char *lpzMsg)
{
	FILE* f=fopen(lpzProfileFile,"at");
	if (f)
	{
		fwrite(lpzMsg,strlen(lpzMsg),1,f);

		fclose(f);
	}
	else
	{
		MPMA::ErrorReport()<<"Unable to open file for CProfiler(error).\n";
	}
}

#endif
