//Initialization and configuration options for different parts of the framework.
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "Setup.h"
#include "DebugRouter.h"
#include <list>
#include <algorithm>

//Setup priorities:
//Elements in the base framework should us a value below 0.
//Elements in the extended framework should use a value above 0.
//During init, they are called lowest to highest.
//During shutdown, they are called highest to lowest.

// -- -- -- Framework Setup -- -- --
namespace
{
	struct PrioritizedCallback
	{
		int priority;
		void (*callback)();
		
		bool operator<(const PrioritizedCallback &other) const
		{
			return priority<other.priority;
		}
	};
	
	std::list<PrioritizedCallback> *internalInitCallbacks;
	std::list<PrioritizedCallback> *internalShutdownCallbacks;
	bool isInited=false;

	//sets up the internal lists for use (will NOT be freed)
	void CheckPreInit()
	{
		if (internalInitCallbacks==0) internalInitCallbacks=new std::list<PrioritizedCallback>;
		if (internalShutdownCallbacks==0) internalShutdownCallbacks=new std::list<PrioritizedCallback>;
	}
}
namespace MPMA
{
	//Internal framework use only: registers init and shutdown callbacks
	void Internal_AddInitCallback(void (*pfunc)(), int priority)
	{
		CheckPreInit();
		PrioritizedCallback pcb;
		pcb.callback=pfunc;
		pcb.priority=priority;
		internalInitCallbacks->push_back(pcb);
	}
	void Internal_AddShutdownCallback(void (*pfunc)(), int priority)
	{
		CheckPreInit();
		PrioritizedCallback pcb;
		pcb.callback=pfunc;
		pcb.priority=priority;
		internalShutdownCallbacks->push_back(pcb);
	}
	
	//performs the callbacks on the inits and shutdowns
	InitAndShutdown::InitAndShutdown()
	{
		CheckPreInit();
		if (isInited)
		{
			ErrorReport()<<"Framework is being initialized, but is already initialized.  Ignoring.\n";
			return;
		}
		
		isInited=true;
		
		//sort the setup list by priority, then call them all
		internalInitCallbacks->sort();
		
		for (std::list<PrioritizedCallback>::iterator i=(*internalInitCallbacks).begin(); i!=(*internalInitCallbacks).end(); ++i)
			(*i).callback();
	}
	InitAndShutdown::~InitAndShutdown()
	{
		if (!isInited)
		{
			ErrorReport()<<"Framework is being shutdown, but is already shutdown.  Ignoring.\n";
			return;
		}
		
		//sort the shutdown list by priority, then call them all
		internalShutdownCallbacks->sort();
		
		for (std::list<PrioritizedCallback>::reverse_iterator i=(*internalShutdownCallbacks).rbegin(); i!=(*internalShutdownCallbacks).rend(); ++i)
			(*i).callback();
		
		//
		isInited=false;
	}
}
