//!\file Thread.h Thread creation and management.
//Luke Lenhart (2007)
//See /docs/License.txt for details on how this code may be used.

#pragma once

#include "Types.h"

namespace MPMA
{
	// -- Thread Class
	
	//!A user-defined parameter passed to the thread procedure.
	union ThreadParam
	{
		//!param as a pointer
		void *ptr;
		//!param as an int
		int val;

		//handy contructors
		inline ThreadParam(void *somePtr): val(0) //!<ctor
			{ ptr=somePtr; }
		inline ThreadParam(int someInt): ptr(0) //!<ctor
			{ val=someInt; }
		inline ThreadParam(const ThreadParam &p)
			{ ptr=p.ptr; val=p.val; }
	};
	
	//!\brief A callback function for the thread's procedure.
	//!This function should monitor the IsEnding() method of the provided Thread class, and terminate if it ever returns true.
	typedef void (*ThreadFunc)(class Thread&, ThreadParam);

	//!Represents a thread.
	class Thread
	{
	public:
		//!Construct a thread and start it.
		Thread(ThreadFunc proc, ThreadParam param);
		
		//!When this object destructs, we block until the thread terminates (if it has not already done so).
		~Thread();
		
		//!Toggles a thread between normal and low priority
		void SetPriorityLow(bool isLowPriority);

		//!Returns true if the thread is still running
		bool IsRunning() const;

		//!Returns true if the thread has been asked to terminate.  This is meant for use by the thread proc itself.
		bool IsEnding() const;

	private:
		//you cannot duplicate a thread
		Thread(const Thread&);
		const Thread& operator=(const Thread&);

		//internal use to store thread state
		void *internal;
	};
	
	
	// -- Thread-related helpers
	
	//!Causes the current thread to block for at least time (in milliseconds).  0 just forces a yield.
	void Sleep(uint time);
}
