//!\file Vary.h A variable type variable.
//useful for simple things where usability is more important than effeciency
//Luke Lenhart (2005)
//See /docs/License.txt for details on how this code may be used.

//Uninitialized type is a Vary::STRING of length 0.
//Converting a string that doesn't contain a number to a number, is just 0.
//A vectors of char's can be used as an input source, and is treated as a string.
//When converting from real to integral types, number is rounded

#pragma once

#include <string>
#include <vector>

namespace MPMA
{
	//!A variable that can have eithar an integral, real, or string type.  It can convert itself between the different types.
	class Vary
	{
	public:
		//!Type of variable.
		enum Type
		{
			INTEGER, //!<signed integer
			REAL,    //!<floating point type
			STRING   //!<character string
		};
	
		//ctors and dtor
		Vary(); //!<ctor
		Vary(const Vary &o); //!<ctor
		inline Vary(const int &ival) {*this=ival;} //!<ctor
		inline Vary(const unsigned int &ival) {*this=ival;} //!<ctor
		inline Vary(const float &rval) {*this=rval;} //!<ctor
		inline Vary(const std::string &sval) {*this=sval;} //!<ctor
		inline Vary(const char *sval) {*this=sval;} //!<ctor
		inline Vary(const std::vector<char> &sval) {*this=sval;} //!<ctor
		~Vary();
	
		//!Retrieves the current type.
		inline const Type GetType() const {return type;}
	
		//!Resets the value of the variable, but not the type.
		void Clear();
		
		//assignment
		const Vary& operator=(const Vary &vval); //!<op=
		const Vary& operator=(const int &ival); //!<op=
		inline const Vary& operator=(const unsigned int &ival) {*this=(int)ival; return *this;} //!<op=
		const Vary& operator=(const float &rval); //!<op=
		const Vary& operator=(const std::string &sval); //!<op=
		const Vary& operator=(const char *sval); //!<op=
		const Vary& operator=(const std::vector<char> &sval); //!<op=
	
		//conversion (non-altering)
		operator const int() const; //!<conversion
		operator const std::string&() const; //!<conversion
		operator const float() const; //!<conversion
		inline const int AsInt() const {return (int)*this;} //!<conversion
		inline const std::string& AsString() const {return (const std::string&)*this;} //!<conversion
		inline const float AsFloat() const {return (float)*this;} //!<conversion
	
		//forced conversion of this instance into another type
		inline void MakeInt() {*this=(int)*this;} //!<Changes the current type.
		inline void MakeReal() {*this=(float)*this;} //!<Changes the current type.
		inline void MakeString() {*this=(const std::string&)*this;} //!<Changes the current type.
	
		//!Addition (for numeric) and concatenation (for string)
		const Vary& operator+=(const Vary &var);
		const Vary& operator+=(const std::string &str);
		const Vary& operator+=(const char *str);
		inline friend const std::string& operator+=(std::string &str, const Vary &v) { return str+=(const std::string&)v; }
			//...letting implicit conversions take care of the rest
	
		//!Addition (for numeric) and concatenation (for string)
		const Vary operator+(const Vary &var) const;
		inline const Vary operator+(const std::string &str) const { return *this+Vary(str); }
		inline const Vary operator+(const char *str) const { return *this+Vary(str); }
		inline friend Vary operator+(const char* str, const Vary &v) { return v+str; }
		inline friend Vary operator+(const std::string &str, const Vary &v) { return v+str; }
			//...letting implicit conversions take care of the rest
	
		//!Increment (valid only on INTEGER and REAL types)
		Vary& operator++(); //pre
		Vary operator++(int); //post
		//!Decrement (valid on all types, on strings it removes the last character (if any))
		Vary& operator--(); //pre
		Vary operator--(int); //post
		
	private:
		Type type; //current type of us
		
		//contents of us (current type has the actual value, the rest are just cache then)
		mutable std::string curStr;
		mutable int curInt;
		mutable float curReal;
	
		mutable unsigned int validBits; //which members are up to date
	};

} //namespace MISC
