//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "../Locks.h"
#include <pthread.h>
#include "../MemMan.h"
#include <errno.h>
#include <time.h>

// -- mutex setup

namespace MPMA
{

MutexLock::MutexLock()
{
	pthread_mutexattr_t mutexattr;
	pthread_mutexattr_settype(&mutexattr, PTHREAD_MUTEX_RECURSIVE_NP);
	
	pthread_mutex_t *m;
	if (mMan.IsReady())
	{
		isMemManaged=true;
		m=new2(pthread_mutex_t, pthread_mutex_t);
	}
	else
	{
		isMemManaged=false;
		m=new pthread_mutex_t;
	}
	pthread_mutex_init(m, &mutexattr);
	pthread_mutexattr_destroy(&mutexattr);
	crit=m;
}

MutexLock::~MutexLock()
{
	pthread_mutex_t* theCrit=(pthread_mutex_t*)crit;
	pthread_mutex_destroy(theCrit);
	
	if (isMemManaged)
		delete2(theCrit);
	else
		delete theCrit;
	
	crit=0;
}


// -- mutex use

TakeMutexLock::TakeMutexLock(MutexLock &critSection):
	crit(critSection)
{
	pthread_mutex_lock((pthread_mutex_t*)crit.crit);
}

TakeMutexLock::~TakeMutexLock()
{
	pthread_mutex_unlock((pthread_mutex_t*)crit.crit);
}


// -- BlockingObject

struct BlockObjData
{
	pthread_cond_t cond;
	pthread_mutex_t condlock;
};

BlockingObject::BlockingObject()
{
	isSet=false;
	pdata=new2(BlockObjData(),BlockObjData);
	BlockObjData *bodata=(BlockObjData*)pdata;

	pthread_cond_init(&bodata->cond, 0);
	pthread_mutex_init(&bodata->condlock, 0);
}

BlockingObject::~BlockingObject()
{
	BlockObjData *bodata=(BlockObjData*)pdata;
	pthread_mutex_destroy(&bodata->condlock);
	pthread_cond_destroy(&bodata->cond);

	delete2(bodata);
}

//Marks the object as cleared, so nothing is blocking on it.
void BlockingObject::Clear()
{
	BlockObjData *bodata=(BlockObjData*)pdata;
	
	TakeSpinLock takeLock(lock);
	pthread_mutex_lock(&bodata->condlock);
	isSet=false;
	
	pthread_cond_broadcast(&bodata->cond);
	pthread_mutex_unlock(&bodata->condlock);
}

//Marks the object as set, so any all Waits will block on it.
void BlockingObject::Set()
{
	TakeSpinLock takeLock(lock);
	isSet=true;
}

//Blocks and waits for as long as the object is set (or if time is non-max, then until that many milliseconds pass).  Returns false only on timeout.
bool BlockingObject::WaitUntilClear(unsigned int timeToWait)
{
	BlockObjData *bodata=(BlockObjData*)pdata;
	
	TakeSpinLock takeLock(lock);
	if (!isSet) return true;

	pthread_mutex_lock(&bodata->condlock);
	
	if (timeToWait==0xffffffff)
	{
		takeLock.Leave();
		pthread_cond_wait(&bodata->cond, &bodata->condlock);
		pthread_mutex_unlock(&bodata->condlock); //since it re-takes it on return
		return true;
	}
	else
	{
		timespec timeout;
		clock_gettime(CLOCK_REALTIME,&timeout);
		timeout.tv_sec+=timeToWait/1000;
		timeout.tv_nsec+=(timeToWait%1000)*1000*1000;
		if (timeout.tv_nsec>1000000000)
		{
			timeout.tv_nsec-=1000000000;
			++timeout.tv_sec;
		}
		
		takeLock.Leave();
		int ret=pthread_cond_timedwait(&bodata->cond, &bodata->condlock, &timeout);
		pthread_mutex_unlock(&bodata->condlock); //since it re-takes it on return
		return ret!=ETIMEDOUT;
	}
}

} //namespace MPMA
