//Debugging helpers
//written by Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "../Debug.h"
#include "../Types.h"
#include "../Locks.h"
#include "../MemMan.h"
#include "evil_windows.h"
#include <dbghelp.h>

namespace
{
#ifdef WIN32_ALLOW_STACK_WALK
	bool symInit=false;
	MPMA::SpinLock *walkLock=0;

	//sets up symbol library for use
	void Win32DbgInitialize()
	{
		walkLock=new2(MPMA::SpinLock(),MPMA::SpinLock);

		if (!SymInitialize(GetCurrentProcess(), 0, true))
		{
			OutputDebugString("SymInitialize failed.  Stack walking will be disabled.\n");
			symInit=false;
			return;
		}

		uint32 opt=SymGetOptions();
		opt|=SYMOPT_UNDNAME;
		SymSetOptions(opt);

		symInit=true;
	}

	//sets up symbol library for use
	void Win32DbgShutdown()
	{
		if (symInit)
		{
			symInit=false;

			SymCleanup(GetCurrentProcess());
		}

		if (walkLock!=0)
		{
			delete2(walkLock);
			walkLock=0;
		}
	}
	
	class AutoWin32DbgInit
	{
	public:
		//hookup init callbacks
		AutoWin32DbgInit()
		{
			MPMA::Internal_AddInitCallback(Win32DbgInitialize,-5000);
			MPMA::Internal_AddShutdownCallback(Win32DbgShutdown,-5000);
		}
	} autoInitDbg;
#endif
}

namespace MPMA
{
	//returns a string representing the call stack (with source locations) for the current thread
#ifdef DEBUG_CALLSTACK_ENABLED
#ifdef WIN32_ALLOW_STACK_WALK
	std::string GetCallStack()
	{
		if (!symInit)
		{
			return "Symbols not initialized.";
		}

		//gather information about the current stack and code location
		uint32 regEIP, regEBP, regESP, addrRet;
		__asm
		{
			//call a function, which pushes the return address onto the stack
			call funcGetEIP
		funcGetEIP:
			//now we'll just pop that return address off and store it... this is EIP.
			pop eax
			mov regEIP, eax
			//get the return address with a similar hack
			pop eax
			mov addrRet, eax
			push eax
			//ESP and EBP we can get directly
			mov regESP, esp
			mov regEBP, ebp
		}

		HANDLE curThread=GetCurrentThread();
		HANDLE curProc=GetCurrentProcess();

		//walk walk walk
		STACKFRAME64 frame;
		memset(&frame, 0, sizeof(STACKFRAME64));
		frame.AddrPC.Mode=AddrModeFlat;
		frame.AddrPC.Offset=regEIP;	
		frame.AddrStack.Mode=AddrModeFlat;
		frame.AddrStack.Offset=regESP;
		frame.AddrFrame.Mode=AddrModeFlat;
		frame.AddrFrame.Offset=regEBP;
		frame.AddrReturn.Mode=AddrModeFlat;
		frame.AddrReturn.Offset=addrRet;

		CONTEXT context;
		memset(&context, 0, sizeof(CONTEXT));

		TakeSpinLock takeLock(*walkLock);

		std::string str="";
		int walkCount=0;
		bool lastAddedAnything=false;
		while (StackWalk64(IMAGE_FILE_MACHINE_I386, curProc, curThread, &frame, &context, 0, SymFunctionTableAccess64, SymGetModuleBase64, 0))
		{
			//this seems to sometimes happen.. just stop if we get a 0 return address
			if (frame.AddrReturn.Offset==0) break;

			if (lastAddedAnything)
				str+="\n";
			lastAddedAnything=false;

			//add what module it's in
			bool wasKernelModule=false;
			IMAGEHLP_MODULE64 module;
			memset(&module, 0, sizeof(module));
			module.SizeOfStruct=sizeof(IMAGEHLP_MODULE64);

			if (SymGetModuleInfo64(curProc, frame.AddrReturn.Offset, &module))
			{
				lastAddedAnything=true;

				str+=module.ModuleName;
				str+="  ";

				if (std::string(module.ModuleName)=="kernel32")
					wasKernelModule=true;
			}

			if (!wasKernelModule) //we get false information about kernel so don't bother
			{
				//add the function name
				union
				{
					//the "name" at the end of this is not a pointer as the documentation says.
					//the string is really just stored after the struct.
					SYMBOL_INFO symInfo; 
					unsigned char symInfoMemory[sizeof(SYMBOL_INFO)+MAX_SYM_NAME+1];
				};
				memset(symInfoMemory, 0, sizeof(symInfoMemory));
				symInfo.SizeOfStruct=sizeof(SYMBOL_INFO);
				symInfo.MaxNameLen=MAX_SYM_NAME;

				if (SymFromAddr(curProc, frame.AddrReturn.Offset, 0, &symInfo))
				{
					lastAddedAnything=true;

					str+=symInfo.Name;
					str+="  ";
				}

				//add the file and line number
				IMAGEHLP_LINE64 line;
				memset(&line, 0, sizeof(IMAGEHLP_LINE64));
				line.SizeOfStruct=sizeof(IMAGEHLP_LINE64);
				uint32 disp=0;

				if (SymGetLineFromAddr64(curProc, frame.AddrReturn.Offset-1, (PDWORD)&disp, &line))
				{
					lastAddedAnything=true;

					str+=line.FileName;
					str+=" line ";

					char tmp[32];
					str+=itoa(line.LineNumber, tmp, 10);
				}
			}

			//
			++walkCount;
			if (walkCount>100) //protect against any sort of silly infinite loop
			{
				break;
			}
		}

		return str+"\n";
	}
#else //WIN32_ALLOW_STACK_WALK
	std::string GetCallStack()
	{
		return "";
	}
#endif //WIN32_ALLOW_STACK_WALK
#endif //DEBUG_CALLSTACK_ENABLED
}
