//!\file Geo.h Vector and matrix primitives and operations.
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

//all angles are in radians

#ifndef GEO_H_INCLUDED
#define GEO_H_INCLUDED

#include <cmath>
#include <string>

#pragma pack(push,4) //align to 4 bytes

#include "base/Types.h"
#include "GeoBases.h"

//!Geometry library
namespace GEO
{
	// -- global constants

	extern const float PI2; //!<2 PI
	extern const float PI; //!<PI
	extern const float FLOAT_TOLERANCE; //!<Acceptable range of closeness to count floats as equal

	
	// -- vector definitions	

	//!A 4d vector or point.
	struct Vector4: public VectorN<4>
	{
		//ctors
		inline Vector4() {} //!<ctor
		inline Vector4(float x, float y, float z, float w) {elem[0]=x; elem[1]=y; elem[2]=z; elem[3]=w;} //!<ctor
		inline Vector4(const float *array): VectorN<4>(array) {} //!<ctor
		inline Vector4(float c): VectorN<4>(c) {} //!<ctor
		inline Vector4(int c): VectorN<4>((float)c) {} //!<ctor
		inline Vector4(const Vector4 &o): VectorN<4>((const float*)&o) {} //!<ctor
		inline Vector4(const VectorN<4> &o): VectorN<4>((const float*)&o) {} //!<ctor
		inline Vector4(const struct Vector3 &o, float w=1); //!<ctor

		//other accessors
		inline float x() const { return elem[0]; } //!<accessor
		inline float& x() { return elem[0]; } //!<accessor
		inline float y() const { return elem[1]; } //!<accessor
		inline float& y() { return elem[1]; } //!<accessor
		inline float z() const { return elem[2]; } //!<accessor
		inline float& z() { return elem[2]; } //!<accessor
		inline float w() const { return elem[3]; } //!<accessor
		inline float& w() { return elem[3]; } //!<accessor
	};


	//!A 3d vector or point.
	struct Vector3: public VectorN<3>
	{
		//ctors
		inline Vector3() {} //!<ctor
		inline Vector3(float x, float y, float z) {elem[0]=x; elem[1]=y; elem[2]=z;} //!<ctor
		inline Vector3(const float *array): VectorN<3>(array) {} //!<ctor
		inline Vector3(float c): VectorN<3>(c) {} //!<ctor
		inline Vector3(int c): VectorN<3>((float)c) {} //!<ctor
		inline Vector3(const Vector3 &o): VectorN<3>((const float*)&o) {} //!<ctor
		inline Vector3(const VectorN<3> &o): VectorN<3>((const float*)&o) {} //!<ctor
		explicit inline Vector3(const Vector4 &o); //!<ctor
		inline Vector3(const struct Vector2 &o, float z=1);
		
		//other accessors
		inline float x() const { return elem[0]; } //!<accessor
		inline float& x() { return elem[0]; } //!<accessor
		inline float y() const { return elem[1]; } //!<accessor
		inline float& y() { return elem[1]; } //!<accessor
		inline float z() const { return elem[2]; } //!<accessor
		inline float& z() { return elem[2]; } //!<accessor
	};

	//!A 2d vector or point.
	struct Vector2: public VectorN<2>
	{
		//ctors
		inline Vector2() {} //!<ctor
		inline Vector2(float x, float y) {elem[0]=x; elem[1]=y;} //!<ctor
		inline Vector2(const float *array): VectorN<2>(array) {} //!<ctor
		inline Vector2(float c): VectorN<2>(c) {} //!<ctor
		inline Vector2(int c): VectorN<2>((float)c) {} //!<ctor
		inline Vector2(const Vector2 &o): VectorN<2>((const float*)&o) {} //!<ctor
		inline Vector2(const VectorN<2> &o): VectorN<2>((const float*)&o) {} //!<ctor
		explicit inline Vector2(const Vector3 &o); //!<ctor
		
		//other accessors
		inline float x() const { return elem[0]; } //!<accessor
		inline float& x() { return elem[0]; } //!<accessor
		inline float y() const { return elem[1]; } //!<accessor
		inline float& y() { return elem[1]; } //!<accessor
	};


	// -- square matrix definitions

	//!A 4x4 matrix.
	struct Matrix4: public MatrixN<4>
	{
		//useful ctors
		inline Matrix4() {} //!<ctor
		inline Matrix4(const float *array): MatrixN<4>(array) {} //!<ctor
		inline Matrix4(float e0,float e1,float e2,float e3,float e4,float e5,float e6,float e7,float e8,float e9,float e10,float e11,float e12,float e13,float e14,float e15); //!<ctor
		inline Matrix4(float c): MatrixN<4>(c) {} //!<ctor
		inline Matrix4(int c): MatrixN<4>((float)c) {} //!<ctor
		inline Matrix4(const Matrix4 &o): MatrixN<4>((const float*)&o) {} //!<ctor
		inline Matrix4(const MatrixN<4> &o): MatrixN<4>((const float*)&o) {} //!<ctor
		inline Matrix4(const struct Matrix3 &o); //!<ctor
	};
	
	//!A 3x3 matrix.
	struct Matrix3: public MatrixN<3>
	{
		//useful ctors
		inline Matrix3() {} //!<ctor
		inline Matrix3(const float *array): MatrixN<3>(array) {} //!<ctor
		inline Matrix3(float e0,float e1,float e2,float e3,float e4,float e5,float e6,float e7,float e8); //!<ctor
		inline Matrix3(float c): MatrixN<3>(c) {} //!<ctor
		inline Matrix3(int c): MatrixN<3>((float)c) {} //!<ctor
		inline Matrix3(const Matrix3 &o): MatrixN<3>((const float*)&o) {} //!<ctor
		inline Matrix3(const MatrixN<3> &o): MatrixN<3>((const float*)&o) {} //!<ctor
		explicit inline Matrix3(const struct Matrix4 &o); //!<ctor
		inline Matrix3(const struct Matrix2 &o); //!<ctor
	};
	
	//!A 2x2 matrix.
	struct Matrix2: public MatrixN<2>
	{
		//useful ctors
		inline Matrix2() {} //!<ctor
		inline Matrix2(const float *array): MatrixN<2>(array) {} //!<ctor
		inline Matrix2(float e0,float e1,float e2,float e3); //!<ctor
		inline Matrix2(float c): MatrixN<2>(c) {} //!<ctor
		inline Matrix2(int c): MatrixN<2>((float)c) {} //!<ctor
		inline Matrix2(const Matrix2 &o): MatrixN<2>((const float*)&o) {} //!<ctor
		inline Matrix2(const MatrixN<2> &o): MatrixN<2>((const float*)&o) {} //!<ctor
		explicit inline Matrix2(const Matrix3 &o); //!<ctor
	};
	
	
	// -- quaternion definition
	
	//TODO
	
	
	// -- generic utilities
	
	inline bool NearEqual(float c0, float c1, float tolerance=FLOAT_TOLERANCE); //!<compare elements for "near equal"
	template <typename VecType>
	inline bool NearEqual(const VecType& v1, const VecType& v2, float tolerance=FLOAT_TOLERANCE); //!<compare elements for "near equal"
	template <typename VecType>
	inline bool NearEqual(const VecType& v, float c, float tolerance=FLOAT_TOLERANCE); //!<compare elements for "near equal"

	inline bool IsFinite(float c); //!<return true if a float is valid
	template <typename VecType>
	inline bool IsFinite(VecType &v); //!<returns true if all elements are valid floats


	// -- Vector functions

	template <typename VecType>
	inline float VecDot(const VecType &v1, const VecType &v2); //!<dot product

	inline struct Vector3 VecCross(const struct Vector3 &v1, const struct Vector3 &v2); //!<cross product
	
	template <typename VecType>
	inline VecType VecNormal(const VecType &v); //!<get the normalized version of a vector
	template <typename VecType>
	inline void VecNormalize(VecType &v); //!<normalize a vector

	template <typename VecType>
	inline float VecAngleBetween(const VecType &l1, const VecType &l2); //!<angle between vectors

	template <typename VecType>
	inline VecType VecReflect(const VecType &vec, const VecType &norm); //!<reflect a vector off of a normal
	
	
	// -- Matrix functions
	
	template <typename MatType>
	MatType MatMul(const MatType &m1, const MatType &m2); //!<multiply two matrices

	float MatDeterminant(const Matrix4 &mat); //!<finds the determinant of a matrix
	float MatDeterminant(const Matrix3 &mat); //!<finds the determinant of a matrix
	inline float MatDeterminant(const Matrix2 &mat); //!<finds the determinant of a matrix

	template <typename MatType>
	MatType MatTranspose(const MatType &mat); //!<finds the transpose of a matrix

	template <typename MatType>
	MatType MatInverse(const MatType &mat); //!<finds the inverse of a matrix (undefined behaviour if inverse does not exist)
	
	inline Matrix3 MatRotateX(float angle); //!<creates a matrix that rotates around the x axis
	inline Matrix3 MatRotateY(float angle); //!<creates a matrix that rotates around the y axis
	inline Matrix3 MatRotateZ(float angle); //!<creates a matrix that rotates around the z axis
	inline Matrix3 MatRotateAxis(const Vector3 &axis, float angle); //!<creates a matrix that rotates around an axis
	
	template <typename VecType>
	inline typename VecType::GreaterMatrixType MatTranslate(const VecType &offset); //!<builds a translation matrix

	Matrix4 MatProjectionFoV(float fov=PI/2, float aspectYdivX=1.0f, float nearZ=1.0f, float farZ=1000.0f); //!<constructs a left-handed perspective projection transform

	Matrix4 MatViewLookAt(const Vector3 &pos, const Vector3 &tar, const Vector3 &vagueUp=Vector3(0,0,1), bool flipVert=false, bool flipHorz=false); //!<constructs a look-at-based left-handed view transformation
	
	
	// -- Mixed functions
	
	template <typename VecType>
	inline VecType TransformVector(const typename VecType::MatrixType &m, const VecType &v); //!<multiply a matrix by a vector
	
}; //namespace GEO

//include template and inline implementations
#ifndef GEO_INCLUDE_INLINE
	#define GEO_INCLUDE_INLINE
	#include "Geo.cpp"
#endif

#pragma pack(pop) //undo alignment change

#endif // GEO_H_INCLUDED
