//Console Unit Test runner
//Luke Lenhart, 2007
//See /docs/License.txt for details on how this code may be used.

#include "UnitTest.h"
#include <list>
#include "base/MemMan.h"
#include "base/Types.h"
#include "base/Profiler.h" //we'll test the profiler on this console itself
#include "base/Setup.h"
#include "base/DebugRouter.h"
#include <iostream>

#define DECLARE_TESTS_CODE
#include "TestList.h"
#undef DECLARE_TESTS_CODE

void MemoryErrorReport(const std::string &msg);

// --
//test runner
int main()
{
	//init framework, and hook up error output to the console
	MPMA::InitAndShutdown scopeAutoInitAndShutdown;
	
	mMan.AddLeakReportCallback(MemoryErrorReport);
	MPMA::RouterOutputStdout *stdReportOutput=new2(MPMA::RouterOutputStdout(), MPMA::RouterOutputStdout);
	MPMA::ErrorReport().AddOutputMethod(stdReportOutput);
	
	//get list of tests to run
	std::list< std::pair<std::string,UnitTest*> > autoBuiltTestList;
	
	ProfileStart("Build Test List");
	#define BUILD_TESTS_LIST
	#include "TestList.h"
	#undef BUILD_TESTS_LIST
	std::list<std::pair<std::string,UnitTest*> > &testList=autoBuiltTestList;
	ProfileStop("Build Test List");

	//lets keep some counts
	uint passed=0;
	uint failed=0;

	//run them!
	for (std::list<std::pair<std::string,UnitTest*> >::iterator i=testList.begin(); i!=testList.end(); ++i)
	{
		ProfileScope("Every Test Run");

		//
		bool didPass=false;
		std::string testName="Unknown Test";
		bool boom=false;
		try
		{
			testName=(*i).first;
			std::cout<<"* Running "<<testName<<"..."<<std::endl;

			ProfileScope(testName.c_str());

			didPass=(*i).second->Run();
		}
		catch(...)
		{
			boom=true;
			std::cout<<"\n* !! Test crashed: "<<testName<<"\n"<<std::endl;
		}
		
		//since each test run has it's own little lifespan, lets flush the error output so that it doesn't mix with the output of a different test
		stdReportOutput->FlushInputSources();
		
		//
		if (didPass)
		{
			std::cout<<"* "<<testName<<" Passed.\n\n";
			++passed;
		}
		else
		{
			if (!boom) std::cout<<"* ! "<<testName<<" Failed.\n\n";
			++failed;
		}
	}

	//cleanup
	ProfileStart("Free Test List");
	while (!testList.empty())
	{
		delete2(testList.front().second);
		testList.front().second=0;
		testList.pop_front();
	}
	ProfileStop("Free Test List");

	//spit out some stats
	std::cout<<"\n* Summary for ";
#ifdef _DEBUG
	std::cout<<"Debug";
#else
	std::cout<<"Release";
#endif
	std::cout<<" build:\n "<<passed<<"   Passed\n "<<failed<<"   Failed\n";

	delete2(stdReportOutput);
	
	return 0;
}

void MemoryErrorReport(const std::string &msg)
{
	std::cout<<msg;
	std::cout.flush();
}
